from dagster_pandas import create_dagster_pandas_dataframe_type, PandasColumn

CurrencyCodeDataFrame = create_dagster_pandas_dataframe_type(
    name="CurrencyCodeDataFrame",
    columns=[
        PandasColumn.datetime_column(
            "partition_key",
            is_required=True,
        ),
        PandasColumn.string_column(
            "entity",
        ),
        PandasColumn.string_column(
            "currency",
        ),
        PandasColumn.string_column(
            "alphabetic_code",
        ),
        PandasColumn.integer_column(
            "numeric_code",
        ),
        PandasColumn.integer_column(
            "minor_unit",
        ),
        PandasColumn.string_column(
            "withdrawal_date",
        ),
    ],
)



from dagster_pandas import create_dagster_pandas_dataframe_type, PandasColumn

CountryDataFrame = create_dagster_pandas_dataframe_type(
    name="CountryDataFrame",
    columns=[
        PandasColumn.integer_column("id"),
        PandasColumn.datetime_column(
            "partition_key",
            is_required=True,
        ),
        PandasColumn.datetime_column(
            "activated",
            is_required=True,
        ),
        PandasColumn.datetime_column(
            "partition_key",
            is_required=True,
        ),
        PandasColumn.datetime_column(
            "created",
            is_required=True,
        ),
        PandasColumn.datetime_column(
            "updated",
            is_required=False,
        ),
        PandasColumn.datetime_column(
            "activated",
            is_required=False,
        ),
        PandasColumn.string_column("name"),
        PandasColumn.string_column("slug"),
        PandasColumn.string_column("description"),
        PandasColumn.string_column("timezones"),
        PandasColumn.string_column("iso_code_2"),
        PandasColumn.string_column("iso_code_3"),
        PandasColumn.string_column("phone_prefix"),
        PandasColumn.string_column("continent"),
    ],
)



from dagster_pandas import create_dagster_pandas_dataframe_type, PandasColumn


# fmt: off
EstablishmentsDataFrame = create_dagster_pandas_dataframe_type(
    name="EstablishmentsDataFrame",
    columns=[
        PandasColumn.datetime_column(
            "partition_key",
            is_required=True,
        ),
        PandasColumn.integer_column(
            "section_number"
        ),
        PandasColumn.string_column("section_name"),
        PandasColumn.string_column("department_number"),
        PandasColumn.string_column("approval_number"),
        PandasColumn.string_column("local_number"),
        PandasColumn.string_column("company_name"),
        PandasColumn.string_column("address"),
        PandasColumn.string_column("postal_code"),
        PandasColumn.string_column("city"),
        PandasColumn.string_column("category"),
        PandasColumn.string_column("associated_activities"),
        PandasColumn.string_column("species"),
    ]
)
# fmt: on



from io import BytesIO

import pandas

from dagster import (
    MemoizableIOManager,
    io_manager,
    MetadataValue,
    Field,
    StringSource,
    get_dagster_logger,
)

from dagster import _check as check


class JsonPartitionedIOManager(MemoizableIOManager):
    def __init__(
        self,
        s3_bucket,
        s3_session,
        s3_prefix=None,
    ):
        self.logger = get_dagster_logger()

        self.bucket = check.str_param(s3_bucket, "s3_bucket")
        self.s3_prefix = check.opt_str_param(s3_prefix, "s3_prefix")
        self.s3 = s3_session
        self.s3.list_objects(Bucket=self.bucket, MaxKeys=1)

    def _get_path(self, context) -> str:
        if context.has_asset_key:
            path = context.get_asset_identifier()
            try:
                del path[path.index(self.bucket)]

            except ValueError:
                bucket_key = self.bucket.replace("-", "_")
                del path[path.index(bucket_key)]

        else:
            path = ["storage", *context.get_identifier()]

        return "/".join(path) + ".json"

    def has_output(self, context):
        key = self._get_path(context)
        return self._has_object(key)

    def _rm_object(self, key):
        check.str_param(key, "key")
        check.param_invariant(len(key) > 0, "key")

        # delete_object wont fail even if the item has been deleted.
        self.s3.delete_object(Bucket=self.bucket, Key=key)

    def _has_object(self, key):
        check.str_param(key, "key")
        check.param_invariant(len(key) > 0, "key")

        try:
            self.s3.get_object(Bucket=self.bucket, Key=key)
            found_object = True
        except self.s3.exceptions.NoSuchKey:
            found_object = False

        return found_object

    def _uri_for_key(self, key):
        check.str_param(key, "key")
        return "s3://" + self.bucket + "/" + "{key}".format(key=key)

    def load_input(self, context) -> pandas.DataFrame:
        if isinstance(context.dagster_type.typing_type, type(None)):
            return None

        key = self._get_path(context)
        if not self._has_object(key):
            context.log.warning(
                f"Object not found from {self._uri_for_key(key)}"
            )
            return None

        context.log.debug(f"Loading S3 object from: {self._uri_for_key(key)}")

        stream_bytes = BytesIO(
            self.s3.get_object(Bucket=self.bucket, Key=key)["Body"].read()
        )

        obj = pandas.read_json(
            stream_bytes,
            orient="records",
            encoding="utf-8",
        )

        return obj

    def handle_output(self, context, obj):
        if isinstance(context.dagster_type.typing_type, type(None)):
            check.invariant(
                obj is None,
                "Output had Nothing type or 'None' annotation, "
                "but handle_output received value "
                f"that was not None and was of type {type(obj)}.",
            )
            return None

        key = self._get_path(context)
        path = self._uri_for_key(key)
        context.log.debug(f"Writing S3 object at: {path}")

        if self._has_object(key):
            context.log.warning(f"Removing existing S3 key: {key}")
            self._rm_object(key)

        if obj is not None:
            pickled_obj = obj.to_json(
                orient="records",
                date_format="iso",
            )

            pickled_obj_bytes = BytesIO(bytes(pickled_obj, "utf-8"))
            self.s3.upload_fileobj(pickled_obj_bytes, self.bucket, key)
            context.add_output_metadata(
                {"uri": MetadataValue.path(path)} | context.metadata
            )

        else:
            context.add_output_metadata(
                {"uri": MetadataValue.path(path), "num_rows": 0}
            )


@io_manager(
    config_schema={
        "s3_bucket": Field(StringSource),
        "s3_prefix": Field(StringSource, is_required=False, default_value=""),
    },
    required_resource_keys={"s3"},
)
def json_partitioned_io_manager(init_context):
    s3_session = init_context.resources.s3
    s3_bucket = init_context.resource_config.get("s3_bucket")
    s3_prefix = init_context.resource_config.get("s3_prefix")

    _json_partitioned_io_manager = JsonPartitionedIOManager(
        s3_bucket, s3_session, s3_prefix=s3_prefix
    )
    return _json_partitioned_io_manager



from dagster_pandas import create_dagster_pandas_dataframe_type, PandasColumn

ContainerCodeDataFrame = create_dagster_pandas_dataframe_type(
    name="ContainerCodeDataFrame",
    columns=[
        PandasColumn.datetime_column(
            "partition_key",
            is_required=True,
        ),
        PandasColumn.string_column(
            "code",
        ),
        PandasColumn.string_column(
            "description",
        ),
        PandasColumn.integer_column(
            "length",
        ),
        PandasColumn.integer_column(
            "height",
        ),
        PandasColumn.string_column(
            "group",
        ),
    ],
)

ContainerGroupDataFrame = create_dagster_pandas_dataframe_type(
    name="ContainerGroupDataFrame",
    columns=[
        PandasColumn.datetime_column(
            "partition_key",
            is_required=True,
        ),
        PandasColumn.string_column(
            "code",
        ),
        PandasColumn.string_column(
            "description",
        ),
    ],
)



from dagster_pandas import create_dagster_pandas_dataframe_type, PandasColumn

LanguageCodeDataFrame = create_dagster_pandas_dataframe_type(
    name="LanguageCodeDataFrame",
    columns=[
        PandasColumn.datetime_column(
            "partition_key",
            is_required=True,
        ),
        PandasColumn.string_column(
            "alpha3_b",
        ),
        PandasColumn.string_column(
            "alpha3_t",
        ),
        PandasColumn.string_column(
            "alpha2",
        ),
        PandasColumn.string_column(
            "english",
        ),
        PandasColumn.string_column(
            "french",
        ),
    ],
)



import pandas
from pandas import DataFrame
from sqlalchemy import create_engine, text
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import ProgrammingError as SQLAlchemyProgrammingError

from dagster import (
    Field,
    StringSource,
    io_manager,
    get_dagster_logger,
    MemoizableIOManager,
)
import dagster._check as check

POSTGRES_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"


class PostgresIOManager(MemoizableIOManager):
    def __init__(
        self,
        username,
        password,
        hostname,
        port,
        dbname,
    ):
        self.logger = get_dagster_logger()

        self.username = username
        self.password = password
        self.hostname = hostname
        self.port = port
        self.dbname = dbname

        self.engine = create_engine(
            f"postgresql+psycopg2://"
            f"{self.username}:{self.password}@{self.hostname}:{self.port}/{self.dbname}"  # noqa: E501
        )
        session_class = sessionmaker(self.engine)
        self.session = session_class()

    def _get_path(self, context) -> str:
        if context.has_asset_key:
            path = context.get_asset_identifier()
            context.log.debug(f"Asset identifier: {path}")

            self.database_name = path[0].replace("-", "_")

            if len(path) == 4:
                table_prefix = path[1].replace("-", "_")
                table_name = path[2].replace("-", "_")

                if table_prefix in table_name:
                    self.schema_name = table_name
                elif "daily_ga" in table_name:
                    self.schema_name = table_name.replace("daily_", "")
                else:
                    self.schema_name = f"{table_prefix}__{table_name}"
            else:
                self.schema_name = path[-2].replace("daily_", "")

        else:
            path = context.get_identifier()
            context.log.debug(f"Identifier: {path}")
            self.database_name = path[0].replace("-", "_")
            self.schema_name = path[-2].replace("daily_", "").replace("-", "_")

        context.log.debug(
            f"Generated path: {self.database_name}.{self.schema_name}"
        )
        return f"{self.database_name}.{self.schema_name}"

    def has_output(self, context):
        key = self._get_path(context)
        self.logger.debug(f"has_output called for {key}")
        return True

    def _rm_object(self, key, obj):

        where_statement = self._get_where_statement(obj)

        sql_statement = f"DELETE FROM {self.schema_name} "
        sql_statement += where_statement
        self.logger.debug(f"Delete on {where_statement} for key {key}")
        self.session.execute(text(sql_statement))
        self.session.commit()

    def _has_object(self, key, obj):

        where_statement = self._get_where_statement(obj)

        sql_statement = f"SELECT * FROM {self.schema_name} "
        sql_statement += where_statement

        try:
            results = self.session.execute(text(sql_statement)).fetchall()
            found_object = bool(len(results))

        except SQLAlchemyProgrammingError as e:
            self.logger.warning(e)
            found_object = False

        return found_object

    def _get_where_statement(self, obj):
        where_statement = "WHERE "

        partition_expr_values = obj[self.partition_expr]
        partition_expr_values = partition_expr_values.drop_duplicates()

        _has_previous_condition = False
        for index, row in partition_expr_values.iterrows():
            if _has_previous_condition:
                where_statement += " OR "

            _has_previous_partition_expr = False

            where_statement += "("

            for expr in self.partition_expr:
                if _has_previous_partition_expr:
                    where_statement += " AND "
                where_statement += f"{expr} = '{row[expr]}'"
                _has_previous_partition_expr = True

            _has_previous_condition = True

            where_statement += ")"

        return where_statement

    def load_input(self, context) -> DataFrame:
        key = self._get_path(context)

        self.logger.debug(
            f"Checking data from {key} for partition {context.partition_key}"
        )

        sql_statement = f"SELECT * FROM {self.schema_name} "

        if context.partition_key:
            sql_statement += f"WHERE partition_key = '{context.partition_key}'"

        obj = pandas.read_sql(sql_statement, con=self.engine.connect())

        return obj

    def handle_output(self, context, obj):
        key = self._get_path(context)

        if isinstance(context.dagster_type.typing_type, type(None)):
            check.invariant(
                obj is None,
                "Output had Nothing type or 'None' annotation, "
                "but handle_output received value "
                f"that was not None and was of type {type(obj)}.",
            )
            return None

        context.log.debug(f"Writing DB object at: {key}")

        if obj is not None:
            if self._has_object(key, obj):
                context.log.warning(f"Removing existing DB entries from {key}.")
                self._rm_object(key, obj)
            obj.to_sql(
                self.schema_name,
                con=self.engine.connect(),
                index=False,
                if_exists="append",
            )
            context.add_output_metadata(
                {
                    "database_name": self.database_name,
                    "schema_name": self.schema_name,
                    "num_rows": len(obj.index),
                }
                | context.metadata
            )

        else:
            context.add_output_metadata(
                {
                    "database_name": self.database_name,
                    "schema_name": self.schema_name,
                    "num_rows": 0,
                }
            )


@io_manager(
    config_schema={
        "username": Field(
            StringSource,
        ),
        "password": Field(
            StringSource,
        ),
        "hostname": Field(
            StringSource,
        ),
        "port": Field(
            StringSource,
        ),
        "dbname": Field(
            StringSource,
        ),
    }
)
def postgres_io_manager(init_context):
    username = init_context.resource_config.get("username")
    password = init_context.resource_config.get("password")
    hostname = init_context.resource_config.get("hostname")
    port = init_context.resource_config.get("port")
    dbname = init_context.resource_config.get("dbname")

    _postgres_io_manager = PostgresIOManager(
        username,
        password,
        hostname,
        port,
        dbname,
    )

    return _postgres_io_manager



from .continents import ContinentDataFrame
from .countries import CountryDataFrame
from .postcodes import PostCodeDataFrame
from .regions import RegionDataFrame
from .subregions import SubRegionDataFrame


__all__ = [
    "ContinentDataFrame",
    "CountryDataFrame",
    "PostCodeDataFrame",
    "RegionDataFrame",
    "SubRegionDataFrame",
]



import ftplib
from hashlib import sha512
from io import BytesIO

from slugify import slugify

from .partitions import get_current_partition


def prepare_object(file_resource):
    file_last_access = None
    file_last_modification = None
    file_size = len(file_resource)

    partition_key = get_current_partition()

    metadata = {
        "last_access": file_last_access,
        "last_modification": file_last_modification,
        "size": file_size,
        "partition": partition_key,
    }

    return partition_key, metadata


def is_file(ftp_client, filename):
    try:
        return ftp_client.size(filename) is not None

    except ftplib.error_perm:
        return False


# fmt: off
def ftp_get(
    context,
    ftp_client,
    path,
    destination,
    partitioned,
    files,
):
    context.log.debug(f"Open {path}")

    ftp_client.cwd(path)
    for filename in ftp_client.nlst():
        context.log.debug(filename)
        remote_path = f"/{path}/{filename}"

        if not is_file(ftp_client, remote_path):
            context.log.debug(f"{filename} skipped...")
            continue

        buffer = BytesIO()
        ftp_client.retrbinary(
            f"RETR {remote_path}",
            callback=buffer.write
        )

        slugify_filename = slugify(filename)

        if len(slugify_filename) <= 4:
            context.log.warning(
                f"Unable to parse filename {filename}"
            )
            continue

        if slugify_filename[-4] != "-":
            context.log.warning(
                f"Unable to parse extension for {filename}"
            )
            continue

        object_filename = ".".join(
            [slugify_filename[:-4], slugify_filename[-4:][1:]]
        )

        if partitioned:
            current_partition = get_current_partition()
            object_name = f"{destination}/{current_partition}/{path}/{object_filename}"  # noqa: E501

        else:
            object_name = f"{destination}/{path}/{object_filename}"

        object_name = object_name.replace("./", "").lower()

        raw_data = buffer.getvalue()
        file_hash = sha512(raw_data).hexdigest()

        partition_key, metadata = prepare_object(
            raw_data,
        )
        metadata["hash"] = file_hash

        file_to_append = {
            "name": object_name,
            "filename": filename,
            "hash": file_hash,
            "partition_key": partition_key,
            "is_new": False,
        }

        object_stat = context.resources.legacy_objects.stat(object_name)
        if object_stat:
            context.log.debug(f"{object_name} already exists")
            # Update object or create a new version

            if (
                metadata["hash"]
                == object_stat.metadata["x-amz-meta-hash"]
            ):
                context.log.debug(
                    f"{filename} and {object_name} seems to be the same"  # noqa: E501
                )
                files.append(file_to_append)
                continue

        context.log.info(
            f"Upload {filename} to s3 as {object_name}..."
        )

        context.resources.legacy_objects.write(
            object_name,
            raw_data,
            "text/csv",
            metadata,
        )
        file_to_append["is_new"] = True
        files.append(file_to_append)
# fmt: on



from datetime import datetime


def cast_dataframe(dataframe, columns):
    for column_name, column_type in columns:
        if column_type == "string":
            dataframe[column_name] = dataframe[column_name].astype(
                "str", errors="ignore"
            )
        elif column_type == "int":
            dataframe[column_name] = dataframe[column_name].astype(
                "int", errors="ignore"
            )
        elif column_type == "float":
            dataframe[column_name] = dataframe[column_name].astype(
                "float", errors="ignore"
            )
        elif column_type == "bool":
            dataframe[column_name] = dataframe[column_name].astype(
                "bool", errors="ignore"
            )
        elif column_type == "datetime":
            dataframe[column_name] = dataframe[column_name].astype(
                "datetime", errors="ignore"
            )
        elif column_type == "timedelta":
            dataframe[column_name] = dataframe[column_name].astype(
                "timedelta", errors="ignore"
            )
        elif column_type == "category":
            dataframe[column_name] = dataframe[column_name].astype(
                "category", errors="ignore"
            )
        else:
            dataframe[column_name] = dataframe[column_name].astype(
                "str", errors="ignore"
            )

    return dataframe


def get_date(value):
    try:
        return datetime.strptime(value, "%d/%m/%Y")
    except ValueError:
        return value


def get_float_to_string(value):
    try:
        return str(int(float(value)))
    except ValueError:
        return value


def get_float_from_string(value):
    try:
        return float(value)
    except ValueError:
        return value


def get_int_from_string(value):
    try:
        return int(value)
    except ValueError:
        return value


def get_string_from_float(value):
    try:
        return str(int(float(value)))
    except ValueError:
        return value


def get_human_input_removed(value):
    if value == "...":
        return ""
    return value



import pandera
from pandera.typing import Series, String, DateTime

from dagster_pandera import pandera_schema_to_dagster_type


# fmt: off
class UnitsOfMeasureDataFrame(pandera.SchemaModel):
    partition_key: Series[DateTime] = pandera.Field(
        nullable=True,
        description="Current partition.",
    )
    level_and_category: Series[String] = pandera.Field(
        nullable=True,
        description="See levels.csv.",
    )
    sector: Series[String] = pandera.Field(
        nullable=True,
        description="Only for Levels 1 and 2.",
    )
    common_code: Series[String] = pandera.Field(
        nullable=True,
        description="UNECE code for the unit of measure.",
    )
    name: Series[String] = pandera.Field(
        nullable=True,
    )
    quantity: Series[String] = pandera.Field(
        nullable=True,
        description="Measured quantity, only for Levels 1 and 2.",
    )
    description: Series[String] = pandera.Field(
        nullable=True,
    )


class UnitsOfMeasureLevelsDataFrame(pandera.SchemaModel):
    partition_key: Series[DateTime] = pandera.Field(
        nullable=True,
        description="Current partition.",
    )
    level_and_category: Series[String] = pandera.Field(
        nullable=True,
        description="See levels.csv.",
    )
    description: Series[String] = pandera.Field(
        nullable=True,
    )


UnitsOfMeasureDataFrameType = pandera_schema_to_dagster_type(
    UnitsOfMeasureDataFrame
)
UnitsOfMeasureLevelsDataFrameType = pandera_schema_to_dagster_type(
    UnitsOfMeasureLevelsDataFrame
)
# fmt: on



import pandera
from pandera.typing import Series, String, Int, Float, DateTime

from dagster_pandera import pandera_schema_to_dagster_type


# fmt: off
class GAEventsDataFrame(pandera.SchemaModel):
    event_category: Series[String] = pandera.Field(
        nullable=True,
        description="The event category.",
    )
    event_action: Series[String] = pandera.Field(
        nullable=True,
        description="Event action.",
    )
    event_label: Series[String] = pandera.Field(
        nullable=True,
        description="Event label.",
    )
    total_events: Series[Int] = pandera.Field(
        nullable=True,
        description="The total number of events for the profile, "
                    "across all categories.",
    )
    unique_events: Series[Int] = pandera.Field(
        nullable=True,
        description="The number of unique events. "
                    "Events in different sessions are counted "
                    "as separate events.",
    )
    event_value: Series[Int] = pandera.Field(
        nullable=True,
        description="Total value of events for the profile.",
    )
    customer_key: Series[String] = pandera.Field(
        nullable=True,
        description="Customer Identification key.",
    )
    customer_name: Series[String] = pandera.Field(
        nullable=True,
        description="Customer name.",
    )
    view_id: Series[Int] = pandera.Field(
        nullable=True,
        description="Google Analytics ViewID.",
    )
    partition_key: Series[DateTime] = pandera.Field(
        nullable=True,
        description="Current partition.",
    )


class GASessionsDataFrame(pandera.SchemaModel):
    continent: Series[String] = pandera.Field(
        nullable=True,
        description="Users' continent, derived from users' IP "
                    "addresses or Geographical IDs.",
    )
    country: Series[String] = pandera.Field(
        nullable=True,
        description="Users' country, derived from their IP "
                    "addresses or Geographical IDs.",
    )
    country_iso_code: Series[String] = pandera.Field(
        nullable=True,
        description="Users' country's ISO code (in ISO-3166-1 alpha-2 format), "
                    "derived from their IP addresses or Geographical IDs. "
                    "For example, BR for Brazil, CA for Canada.",
    )
    region: Series[String] = pandera.Field(
        nullable=True,
        description="Users' region, derived from their IP "
                    "addresses or Geographical IDs. "
                    "In U.S., a region is a state, New York, for example.",
    )
    region_iso_code: Series[String] = pandera.Field(
        nullable=True,
        description="Users' region ISO code in ISO-3166-2 format, "
                    "derived from their IP addresses or Geographical IDs.",
    )
    city: Series[String] = pandera.Field(
        nullable=True,
        description="Users' city, derived from their IP "
                    "addresses or Geographical IDs.",
    )
    page_path: Series[String] = pandera.Field(
        nullable=True,
        description="A page on the website specified by path and/or "
                    "query parameters. Use this with hostname "
                    "to get the page's full URL.",
    )
    sessions: Series[Int] = pandera.Field(
        nullable=True,
        description="The total number of sessions.",
    )
    bounces: Series[Int] = pandera.Field(
        nullable=True,
        description="The total number of single page "
                    "(or single interaction hit) sessions for the property.",
    )
    sessions_per_user: Series[Float] = pandera.Field(
        nullable=True,
        description="The total number of sessions divided by "
                    "the total number of users.",
    )
    session_duration: Series[Float] = pandera.Field(
        nullable=True,
        description="Total duration (in seconds) of users' sessions.",
    )
    page_load_time: Series[Int] = pandera.Field(
        nullable=True,
        description="Total time (in milliseconds), from pageview initiation "
                    "(e.g., a click on a page link) to page load completion "
                    "in the browser, the pages in the sample set take to load.",
    )
    customer_key: Series[String] = pandera.Field(
        nullable=True,
        description="Customer Identification key.",
    )
    customer_name: Series[String] = pandera.Field(
        nullable=True,
        description="Customer name.",
    )
    view_id: Series[Int] = pandera.Field(
        nullable=True,
        description="Google Analytics ViewID.",
    )
    partition_key: Series[DateTime] = pandera.Field(
        nullable=True,
        description="Current partition.",
    )


class GAUsersDataFrame(pandera.SchemaModel):
    continent: Series[String] = pandera.Field(
        nullable=True,
        description="Users' continent, derived from users' IP "
                    "addresses or Geographical IDs.",
    )
    country: Series[String] = pandera.Field(
        nullable=True,
        description="Users' country, derived from their IP "
                    "addresses or Geographical IDs.",
    )
    country_iso_code: Series[String] = pandera.Field(
        nullable=True,
        description="Users' country's ISO code (in ISO-3166-1 alpha-2 format), "
                    "derived from their IP addresses or Geographical IDs. "
                    "For example, BR for Brazil, CA for Canada.",
    )
    region: Series[String] = pandera.Field(
        nullable=True,
        description="Users' region, derived from their IP "
                    "addresses or Geographical IDs. "
                    "In U.S., a region is a state, New York, for example.",
    )
    region_iso_code: Series[String] = pandera.Field(
        nullable=True,
        description="Users' region ISO code in ISO-3166-2 format, "
                    "derived from their IP addresses or Geographical IDs.",
    )
    city: Series[String] = pandera.Field(
        nullable=True,
        description="Users' city, derived from their IP "
                    "addresses or Geographical IDs.",
    )
    page_path: Series[String] = pandera.Field(
        nullable=True,
        description="A page on the website specified by path and/or "
                    "query parameters. Use this with hostname "
                    "to get the page's full URL.",
    )
    users: Series[Int] = pandera.Field(
        nullable=True,
        description="The total number of users for the requested time period.",
    )
    new_users: Series[Int] = pandera.Field(
        nullable=True,
        description="The number of sessions marked as a user's first sessions.",
    )
    percent_new_sessions: Series[Float] = pandera.Field(
        nullable=True,
        description="The percentage of sessions by users who had never "
                    "visited the property before.",
    )
    sessions_per_user: Series[Float] = pandera.Field(
        nullable=True,
        description="The total number of sessions divided by "
                    "the total number of users.",
    )
    session_duration: Series[Float] = pandera.Field(
        nullable=True,
        description="Total duration (in seconds) of users' sessions.",
    )
    page_load_time: Series[Int] = pandera.Field(
        nullable=True,
        description="Total time (in milliseconds), from pageview initiation "
                    "(e.g., a click on a page link) to page load completion "
                    "in the browser, the pages in the sample set take to load.",
    )
    customer_key: Series[String] = pandera.Field(
        nullable=True,
        description="Customer Identification key.",
    )
    customer_name: Series[String] = pandera.Field(
        nullable=True,
        description="Customer name.",
    )
    view_id: Series[Int] = pandera.Field(
        nullable=True,
        description="Google Analytics ViewID.",
    )
    partition_key: Series[DateTime] = pandera.Field(
        nullable=True,
        description="Current partition.",
    )


GAEventsDataFrameType = pandera_schema_to_dagster_type(GAEventsDataFrame)
GASessionsDataFrameType = pandera_schema_to_dagster_type(GASessionsDataFrame)
GAUsersDataFrameType = pandera_schema_to_dagster_type(GAUsersDataFrame)
# fmt: on



from io import BytesIO
from os import environ

from minio import Minio
from minio.commonconfig import CopySource
from minio.error import S3Error

from dagster import get_dagster_logger


class ObjectStorage:
    client = None

    hostname = None
    access_key = None
    secret_key = None

    bucket = None

    def __init__(self):
        self.logger = get_dagster_logger()

        self.hostname = environ.get("MINIO_SERVER_URL")
        self.access_key = environ.get("MINIO_ROOT_USER")
        self.secret_key = environ.get("MINIO_ROOT_PASSWORD")

        # fmt: off
        if self.hostname:
            self.hostname = (
                self.hostname.replace("http://", "").replace("https://", "")
            )
        # fmt: on

        self.client = Minio(
            self.hostname,
            secure=False,
            access_key=self.access_key,
            secret_key=self.secret_key,
        )

    def read(self, object_name):
        try:
            response = self.client.get_object(
                bucket_name=self.bucket,
                object_name=object_name,
            )
            content = BytesIO(response.read())
            response.close()
            response.release_conn()

        except S3Error as e:
            self.logger.error(f"Caught {e} for {self.bucket}/{object_name}")
            raise e

        return content

    def write(self, object_name, data, content_type, metadata=None):
        data_buffer = BytesIO(data)

        self.client.put_object(
            bucket_name=self.bucket,
            object_name=object_name,
            data=data_buffer,
            length=len(data),
            content_type=content_type,
            metadata=metadata,
        )

    def stat(self, object_name):
        try:
            return self.client.stat_object(
                bucket_name=self.bucket,
                object_name=object_name,
            )

        except S3Error as e:
            self.logger.error(f"Caught {e} for {self.bucket}/{object_name}")
            return False

    def scan(self, prefix, recursive=True):
        return self.client.list_objects(
            bucket_name=self.bucket,
            prefix=prefix,
            recursive=recursive,
        )

    def copy(self, from_path, to_bucket, to_path):
        self.client.copy_object(
            to_bucket, to_path, CopySource(self.bucket, from_path)
        )
